package dev.micle.xptools.events.common;

import dev.micle.xptools.XpTools;
import dev.micle.xptools.config.Config;
import dev.micle.xptools.config.OperationCache;
import dev.micle.xptools.config.OperationItem;
import net.minecraft.resources.ResourceLocation;
import net.minecraft.tags.TagKey;
import net.minecraft.world.level.block.Block;
import net.minecraftforge.event.level.BlockEvent;
import net.minecraftforge.eventbus.api.SubscribeEvent;
import net.minecraftforge.registries.ForgeRegistries;

import java.time.Duration;
import java.time.Instant;
import java.util.ArrayList;
import java.util.Comparator;
import java.util.List;
import java.util.concurrent.ThreadLocalRandom;

public class OnBlockBreakEventHandler {
    @SubscribeEvent
    public void OnBlockBreakEvent(BlockEvent.BreakEvent event) {
        float xpToDrop = event.getExpToDrop();

        // Get Block id
        ResourceLocation block_rl = ForgeRegistries.BLOCKS.getKey(event.getState().getBlock());
        String block_id = "";
        if (block_rl != null) {
            block_id = block_rl.toString();
        }

        // Collect operations
        Instant start = Instant.now();
        List<OperationItem> operations = OperationCache.getBlockBreakCacheEntry(block_id);
        if (operations != null) {
            XpTools.LOGGER.debug("Got operations from cache for {} [time: {}]", block_id, Duration.between(start, Instant.now()).toNanos());
        } else {
            operations = new ArrayList<>();

            // Collect operations on relevant block_id
            if (!block_id.isEmpty()) {
                for (OperationItem operationItem : Config.Server.blockBreakOperationItems) {
                    if (!operationItem.isTag() && operationItem.getId().equals(block_id)) {
                        operations.add(operationItem);
                    }
                }
            }

            // Collect operations on relevant tag_id
            for (TagKey<Block> tagKey : event.getState().getTags().toList()) {
                String tag_id = tagKey.location().toString();
                for (OperationItem operationItem : Config.Server.blockBreakOperationItems) {
                    if (operationItem.isTag() && operationItem.getId().equals(tag_id)) {
                        operations.add(operationItem);
                    }
                }
            }

            // Sort operations based on priority
            operations.sort(Comparator.comparingInt(OperationItem::getPriority));

            // Remove any operations after last operation
            for (OperationItem operationItem : operations) {
                if (operationItem.isLast()) {
                    operations = operations.subList(0, operations.indexOf(operationItem) + 1);
                    break;
                }
            }

            XpTools.LOGGER.debug("Calculated operations for {} [time: {}]", block_id, Duration.between(start, Instant.now()).toNanos());

            // Save operations to cache
            OperationCache.addBlockBreakCacheEntry(block_id, operations);
        }

        // Add global operations before all others
        operations.addAll(0, Config.Server.blockBreakGlobalOperationItems);

        // Apply operations to xp drops
        for (OperationItem operation : operations) {
            // Calculate operation value
            float opValue = (operation.getMin() == operation.getMax()) ?
                    operation.getMin() :
                    ThreadLocalRandom.current().nextFloat(operation.getMin(), operation.getMax());

            // Apply operation
            switch (operation.getType()) {
                case SET:
                    xpToDrop = opValue;
                    break;
                case ADD:
                    xpToDrop += opValue;
                    break;
                case SUBTRACT:
                    xpToDrop -= opValue;
                    break;
                case MULTIPLY:
                    xpToDrop *= opValue;
                    break;
                case DIVIDE:
                    xpToDrop /= opValue;
                    break;
            }

            // Stop if this is the last operation
            if (operation.isLast()) {
                break;
            }
        }
        XpTools.LOGGER.debug(operations.toString());
        XpTools.LOGGER.debug(String.valueOf(xpToDrop));

        // Apply xp drop
        event.setExpToDrop((int)xpToDrop);
    }
}
